/* jshint globalstrict: true, es3: true */
/* globals require: false, exports: false */
'use strict';
var AqlError = require('./errors').AqlError,
  keywords = require('./assumptions').keywords;

function wrapAQL(expr) {
  if (
    expr instanceof Operation ||
    expr instanceof Statement ||
    expr instanceof PartialStatement
  ) {
    return '(' + expr.toAQL() + ')';
  }
  return expr.toAQL();
}

function autoCastToken(token) {
  var match;
  if (token === null || token === undefined) {
    return new NullLiteral();
  }
  if (token instanceof Expression || token instanceof PartialStatement) {
    return token;
  }
  if (typeof token === 'number' && token === token && token !== Infinity && token !== -Infinity) {
    if (Math.floor(token) === token) {
      return new IntegerLiteral(token);
    }
    return new NumberLiteral(token);
  }
  if (typeof token === 'boolean') {
    return new BooleanLiteral(token);
  }
  if (typeof token === 'string') {
    if (token.match(/^[-+]?[0-9]+(\.[0-9]+)?$/)) {
      return autoCastToken(Number(token));
    }
    if (token.charAt(0) === '"') {
      return new StringLiteral(JSON.parse(token));
    }
    match = token.match(/^([0-9]+)\.\.([0-9]+)$/);
    if (match) {
      return new RangeExpression(Number(match[1]), Number(match[2]));
    }
    if (token.match(Identifier.re)) {
      return new Identifier(token);
    }
    return new SimpleReference(token);
  }
  if (typeof token === 'object') {
    if (token.constructor && token.constructor.name === 'ArangoCollection') {
      return new Identifier(token.name());
    }
    if (Object.prototype.toString.call(token) === '[object Array]') {
      return new ListLiteral(token);
    }
    return new ObjectLiteral(token);
  }
  throw new AqlError('Invalid AQL value: (' + (typeof token) + ') ' + token);
}

function Expression() {}
function Operation() {}
Operation.prototype = new Expression();
Operation.prototype.constructor = Operation;

function RawExpression(value) {
  if (value && value instanceof RawExpression) {value = value.value;}
  this.value = value;
}
RawExpression.prototype = new Expression();
RawExpression.prototype.constructor = RawExpression;
RawExpression.prototype.toAQL = function () {return String(this.value);};

function NullLiteral(value) {
  if (value && value instanceof NullLiteral) {value = value.value;}
  if (value !== null && value !== undefined) {
    throw new AqlError('Expected value to be null: ' + value);
  }
  this.value = value;
}
NullLiteral.prototype = new Expression();
NullLiteral.prototype.constructor = NullLiteral;
NullLiteral.prototype.toAQL = function () {return 'null';};

function BooleanLiteral(value) {
  if (value && value instanceof BooleanLiteral) {value = value.value;}
  this.value = Boolean(value);
}
BooleanLiteral.prototype = new Expression();
BooleanLiteral.prototype.constructor = BooleanLiteral;
BooleanLiteral.prototype.toAQL = function () {return String(this.value);};

function NumberLiteral(value) {
  if (value && (
    value instanceof NumberLiteral ||
    value instanceof IntegerLiteral
  )) {value = value.value;}
  this.value = Number(value);
  if (this.value !== this.value || this.value === Infinity) {
    throw new AqlError('Expected value to be a finite number: ' + value);
  }
}
NumberLiteral.prototype = new Expression();
NumberLiteral.prototype.constructor = NumberLiteral;
NumberLiteral.prototype.toAQL = function () {return String(this.value);};

function IntegerLiteral(value) {
  if (value && (
    value instanceof NumberLiteral ||
    value instanceof IntegerLiteral
  )) {value = value.value;}
  this.value = Number(value);
  if (this.value !== this.value || this.value === Infinity || Math.floor(this.value) !== this.value) {
    throw new AqlError('Expected value to be a finite integer: ' + value);
  }
}
IntegerLiteral.prototype = new Expression();
IntegerLiteral.prototype.constructor = IntegerLiteral;
IntegerLiteral.prototype.toAQL = function () {return String(this.value);};

function StringLiteral(value) {
  if (value && value instanceof StringLiteral) {value = value.value;}
  if (value && typeof value.toAQL === 'function') {value = value.toAQL();}
  this.value = String(value);
}
StringLiteral.prototype = new Expression();
StringLiteral.prototype.constructor = StringLiteral;
StringLiteral.prototype.toAQL = function () {return JSON.stringify(this.value);};

function ListLiteral(value) {
  if (value && value instanceof ListLiteral) {value = value.value;}
  if (!value || Object.prototype.toString.call(value) !== '[object Array]') {
    throw new AqlError('Expected value to be an array: ' + value);
  }
  this.value = [];
  for (var i = 0; i < value.length; i++) {
    this.value[i] = autoCastToken(value[i]);
  }
}
ListLiteral.prototype = new Expression();
ListLiteral.prototype.constructor = ListLiteral;
ListLiteral.prototype.toAQL = function () {
  var value = [], i;
  for (i = 0; i < this.value.length; i++) {
    value.push(wrapAQL(this.value[i]));
  }
  return '[' + value.join(', ') + ']';
};

function ObjectLiteral(value) {
  if (value && value instanceof ObjectLiteral) {value = value.value;}
  if (!value || typeof value !== 'object') {
    throw new AqlError('Expected value to be an object: ' + value);
  }
  this.value = {};
  for (var key in value) {
    if (value.hasOwnProperty(key)) {
      this.value[key] = autoCastToken(value[key]);
    }
  }
}
ObjectLiteral.prototype = new Expression();
ObjectLiteral.prototype.constructor = ObjectLiteral;
ObjectLiteral.prototype.toAQL = function () {
  var items = [], key;
  for (key in this.value) {
    if (this.value.hasOwnProperty(key)) {
      if (key.match(Identifier.re) || key === String(Number(key))) {
        items.push(key + ': ' + wrapAQL(this.value[key]));
      } else {
        items.push(JSON.stringify(key) + ': ' + wrapAQL(this.value[key]));
      }
    }
  }
  return '{' + items.join(', ') + '}';
};

function RangeExpression(start, end) {
  this.start = autoCastToken(start);
  this.end = autoCastToken(end);
}
RangeExpression.prototype = new Expression();
RangeExpression.prototype.constructor = RangeExpression;
RangeExpression.prototype.toAQL = function () {
  return wrapAQL(this.start) + '..' + wrapAQL(this.end);
};

function PropertyAccess(obj, key) {
  this.obj = autoCastToken(obj);
  this.key = autoCastToken(key);
}
PropertyAccess.prototype = new Expression();
PropertyAccess.prototype.constructor = PropertyAccess;
PropertyAccess.prototype.toAQL = function () {
  return wrapAQL(this.obj) + '[' + wrapAQL(this.key) + ']';
};

function Keyword(value) {
  if (value && value instanceof Keyword) {value = value.value;}
  if (!value || typeof value !== 'string') {
    throw new AqlError('Expected value to be a string: ' + value);
  }
  if (!value.match(Keyword.re)) {
    throw new AqlError('Not a valid keyword: ' + value);
  }
  this.value = value;
}
Keyword.re = /^[_a-z][_0-9a-z]*$/i;
Keyword.prototype = new Expression();
Keyword.prototype.constructor = Keyword;
Keyword.prototype.toAQL = function () {
  return String(this.value).toUpperCase();
};

function Identifier(value) {
  if (value && value instanceof Identifier) {value = value.value;}
  if (!value || typeof value !== 'string') {
    throw new AqlError('Expected value to be a string: ' + value);
  }
  if (!value.match(Identifier.re)) {
    throw new AqlError('Not a valid identifier: ' + value);
  }
  this.value = value;
}
Identifier.re = /^[_a-z][_0-9a-z]*$/i;
Identifier.prototype = new Expression();
Identifier.prototype.constructor = Identifier;
Identifier.prototype.toAQL = function () {
  var value = String(this.value);
  if (keywords.indexOf(value.toLowerCase()) === -1) {
    return value;
  }
  return '`' + value + '`';
};

function SimpleReference(value) {
  if (value && value instanceof SimpleReference) {value = value.value;}
  if (!value || typeof value !== 'string') {
    throw new AqlError('Expected value to be a string: ' + value);
  }
  if (!value.match(SimpleReference.re)) {
    throw new AqlError('Not a valid simple reference: ' + value);
  }
  this.value = value;
}
SimpleReference.re = /^@{0,2}[_a-z][_0-9a-z]*(\.[_a-z][_0-9a-z]*|\[\*\])*$/i;
SimpleReference.prototype = new Expression();
SimpleReference.prototype.constructor = SimpleReference;
SimpleReference.prototype.toAQL = function () {
  var value = String(this.value),
    tokens = value.split('.'),
    i;
  for (i = 0; i < tokens.length; i++) {
    if (keywords.indexOf(tokens[i]) !== -1) {
      tokens[i] = '`' + tokens[i] + '`';
    }
  }
  return tokens.join('.');
};

function UnaryOperation(operator, value) {
  if (!operator || typeof operator !== 'string') {
    throw new AqlError('Expected operator to be a string: ' + operator);
  }
  this.operator = operator;
  this.value = autoCastToken(value);
}
UnaryOperation.prototype = new Expression();
UnaryOperation.prototype.constructor = UnaryOperation;
UnaryOperation.prototype.toAQL = function () {
  return this.operator + wrapAQL(this.value);
};

function BinaryOperation(operator, value1, value2) {
  if (!operator || typeof operator !== 'string') {
    throw new AqlError('Expected operator to be a string: ' + operator);
  }
  this.operator = operator;
  this.value1 = autoCastToken(value1);
  this.value2 = autoCastToken(value2);
}
BinaryOperation.prototype = new Operation();
BinaryOperation.prototype.constructor = BinaryOperation;
BinaryOperation.prototype.toAQL = function () {
  return [wrapAQL(this.value1), this.operator, wrapAQL(this.value2)].join(' ');
};

function TernaryOperation(operator1, operator2, value1, value2, value3) {
  if (!operator1 || typeof operator1 !== 'string') {
    throw new AqlError('Expected operator 1 to be a string: ' + operator1);
  }
  if (!operator2 || typeof operator2 !== 'string') {
    throw new AqlError('Expected operator 2 to be a string: ' + operator2);
  }
  this.operator1 = operator1;
  this.operator2 = operator2;
  this.value1 = autoCastToken(value1);
  this.value2 = autoCastToken(value2);
  this.value3 = autoCastToken(value3);
}
TernaryOperation.prototype = new Operation();
TernaryOperation.prototype.constructor = TernaryOperation;
TernaryOperation.prototype.toAQL = function () {
  return [
    wrapAQL(this.value1),
    this.operator1,
    wrapAQL(this.value2),
    this.operator2,
    wrapAQL(this.value3)
  ].join(' ');
};

function NAryOperation(operator, values) {
  if (!operator || typeof operator !== 'string') {
    throw new AqlError('Expected operator to be a string: ' + operator);
  }
  this.operator = operator;
  this.values = [];
  for (var i = 0; i < values.length; i++) {
    this.values.push(autoCastToken(values[i]));
  }
}
NAryOperation.prototype = new Operation();
NAryOperation.prototype.constructor = NAryOperation;
NAryOperation.prototype.toAQL = function () {
  var values = [], i;
  for (i = 0; i < this.values.length; i++) {
    values.push(wrapAQL(this.values[i]));
  }
  return values.join(' ' + this.operator + ' ');
};

function FunctionCall(functionName, args) {
  if (!functionName || typeof functionName !== 'string') {
    throw new AqlError('Expected function name to be a string: ' + functionName);
  }
  if (!functionName.match(FunctionCall.re)) {
    throw new AqlError('Not a valid function name: ' + functionName);
  }
  if (args && Object.prototype.toString.call(args) !== '[object Array]') {
    throw new AqlError('Expected arguments to be an array: ' + args);
  }
  this.functionName = functionName;
  this.args = [];
  if (args) {
    for (var i = 0; i < args.length; i++) {
      this.args[i] = autoCastToken(args[i]);
    }
  }
}
FunctionCall.re = /^[_a-z][_0-9a-z]*(::[_a-z][_0-9a-z]*)*$/i;
FunctionCall.prototype = new Expression();
FunctionCall.prototype.constructor = FunctionCall;
FunctionCall.prototype.toAQL = function () {
  var args = [], i;
  for (i = 0; i < this.args.length; i++) {
    args.push(wrapAQL(this.args[i]));
  }
  return this.functionName + '(' + args.join(', ') + ')';
};

function PartialStatement() {}
PartialStatement.prototype.for_ = function (varname) {
  var self = this, inFn;
  inFn = function (expr) {
    // assert expr is an expression
    return new ForExpression(self, varname, expr);
  };
  return {'in': inFn, in_: inFn};
};
PartialStatement.prototype.filter = function (expr) {return new FilterExpression(this, expr);};
PartialStatement.prototype.let_ = function (varname, expr) {
  if (expr === undefined) {
    return new LetExpression(this, varname);
  }
  return new LetExpression(this, [[varname, expr]]);
};
PartialStatement.prototype.collect = function (varname, expr) {
  if (expr === undefined) {
    return new CollectExpression(this, varname);
  }
  return new CollectExpression(this, [[varname, expr]]);
};
PartialStatement.prototype.sort = function () {
  var args = Array.prototype.slice.call(arguments);
  return new SortExpression(this, args);
};
PartialStatement.prototype.limit = function (x, y) {return new LimitExpression(this, x, y);};
PartialStatement.prototype.return_ = function (x) {return new ReturnExpression(this, x);};
PartialStatement.prototype.remove = function (expr) {
  var self = this, inFn;
  inFn = function (collection) {
    return new RemoveExpression(self, expr, collection);
  };
  return {into: inFn, 'in': inFn, in_: inFn};
};
PartialStatement.prototype.insert = function (expr) {
  var self = this, inFn;
  inFn = function (collection) {
    return new InsertExpression(self, expr, collection);
  };
  return {into: inFn, 'in': inFn, in_: inFn};
};
PartialStatement.prototype.update = function (expr) {
  var self = this, withFn, inFn;
  withFn = function (withExpr) {
    var inFn = function (collection) {
      return new UpdateExpression(self, expr, withExpr, collection);
    };
    return {into: inFn, 'in': inFn, in_: inFn};
  };
  inFn = function (collection) {
    return new ReplaceExpression(self, expr, undefined, collection);
  };
  return {'with': withFn, with_: withFn, into: inFn, 'in': inFn, in_: inFn};
};
PartialStatement.prototype.replace = function (expr) {
  var self = this, withFn, inFn;
  withFn = function (withExpr) {
    var inFn = function (collection) {
      return new ReplaceExpression(self, expr, withExpr, collection);
    };
    return {into: inFn, 'in': inFn, in_: inFn};
  };
  inFn = function (collection) {
    return new ReplaceExpression(self, expr, undefined, collection);
  };
  return {'with': withFn, with_: withFn, into: inFn, 'in': inFn, in_: inFn};
};

PartialStatement.prototype['for'] = PartialStatement.prototype.for_;
PartialStatement.prototype['let'] = PartialStatement.prototype.let_;
PartialStatement.prototype['return'] = PartialStatement.prototype.return_;

function Definitions(dfns) {
  if (dfns instanceof Definitions) {
    dfns = dfns.dfns;
  }
  this.dfns = [];
  if (!dfns || typeof dfns !== 'object') {
    throw new AqlError('Expected definitions to be an object');
  }
  if (Object.prototype.toString.call(dfns) === '[object Array]') {
    for (var i = 0; i < dfns.length; i++) {
      if (Object.prototype.toString.call(dfns[i]) !== '[object Array]' || dfns[i].length !== 2) {
        throw new AqlError('Expected definitions[' + i + '] to be a tuple');
      }
      this.dfns.push([new Identifier(dfns[i][0]), autoCastToken(dfns[i][1])]);
    }
  } else {
    for (var key in dfns) {
      if (dfns.hasOwnProperty(key)) {
        this.dfns.push([new Identifier(key), autoCastToken(dfns[key])]);
      }
    }
  }
  if (this.dfns.length === 0) {
    throw new AqlError('Expected definitions not to be empty');
  }
}
Definitions.prototype.toAQL = function () {
  var dfns = [];
  for (var i = 0; i < this.dfns.length; i++) {
    dfns.push(this.dfns[i][0].toAQL() + ' = ' + wrapAQL(this.dfns[i][1]));
  }
  return dfns.join(', ');
};

function ForExpression(prev, varname, expr) {
  this.prev = prev;
  this.varname = new Identifier(varname);
  this.expr = autoCastToken(expr);
}
ForExpression.prototype = new PartialStatement();
ForExpression.prototype.constructor = ForExpression;
ForExpression.prototype.toAQL = function () {
  return (
    (this.prev ? this.prev.toAQL() + ' ' : '') +
    'FOR ' + wrapAQL(this.varname) +
    ' IN ' + wrapAQL(this.expr)
  );
};

function FilterExpression(prev, expr) {
  this.prev = prev;
  this.expr = autoCastToken(expr);
}
FilterExpression.prototype = new PartialStatement();
FilterExpression.prototype.constructor = FilterExpression;
FilterExpression.prototype.toAQL = function () {
  return (
    (this.prev ? this.prev.toAQL() + ' ' : '') +
    'FILTER ' + wrapAQL(this.expr)
  );
};

function LetExpression(prev, dfns) {
  this.prev = prev;
  this.dfns = new Definitions(dfns);
}
LetExpression.prototype = new PartialStatement();
LetExpression.prototype.constructor = LetExpression;
LetExpression.prototype.toAQL = function () {
  return (
    (this.prev ? this.prev.toAQL() + ' ' : '') +
    'LET ' + this.dfns.toAQL()
  );
};

function CollectExpression(prev, dfns) {
  this.prev = prev;
  this.dfns = new Definitions(dfns);
}
CollectExpression.prototype = new PartialStatement();
CollectExpression.prototype.constructor = CollectExpression;
CollectExpression.prototype.into = function (varname) {
  return new CollectIntoExpression(this.prev, this.dfns, varname);
};
CollectExpression.prototype.toAQL = function () {
  return (
    (this.prev ? this.prev.toAQL() + ' ' : '') +
    'COLLECT ' + this.dfns.toAQL()
  );
};

function CollectIntoExpression(prev, dfns, varname) {
  this.prev = prev;
  this.dfns = new Definitions(dfns);
  this.varname = new Identifier(varname);
}
CollectIntoExpression.prototype = new PartialStatement();
CollectIntoExpression.prototype.constructor = CollectIntoExpression;
CollectIntoExpression.prototype.toAQL = function () {
  return (
    (this.prev ? this.prev.toAQL() + ' ' : '') +
    'COLLECT ' + this.dfns.toAQL() +
    ' INTO ' + this.varname.toAQL()
  );
};

function SortExpression(prev, args) {
  if (!args || Object.prototype.toString.call(args) !== '[object Array]') {
    throw new AqlError('Expected sort list to be an array: ' + args);
  }
  if (!args.length) {
    throw new AqlError('Expected sort list not to be empty: ' + args);
  }
  this.prev = prev;
  this.args = [];
  var allowKeyword = false, i, value;
  for (i = 0; i < args.length; i++) {
    value = args[i];
    if (!allowKeyword && value) {
      if (value instanceof Keyword || (
        typeof value === 'string' && SortExpression.keywords.indexOf(value.toUpperCase()) !== -1
      )) {
        throw new AqlError('Unexpected keyword ' + value.toString() + ' at offset ' + i);
      }
    }
    if (typeof value === 'string' && SortExpression.keywords.indexOf(value.toUpperCase()) !== -1) {
      this.args[i] = new Keyword(value);
      allowKeyword = false;
    } else {
      this.args[i] = autoCastToken(value);
      allowKeyword = true;
    }
  }
}
SortExpression.keywords = ['ASC', 'DESC'];
SortExpression.prototype = new PartialStatement();
SortExpression.prototype.constructor = SortExpression;
SortExpression.prototype.toAQL = function () {
  var args = [], j = 0;
  for (var i = 0; i < this.args.length; i++) {
    if (this.args[i] instanceof Keyword) {
      args[j] += ' ' + this.args[i].toAQL();
    } else {
      j = args.push(wrapAQL(this.args[i])) - 1;
    }
  }
  return (
    (this.prev ? this.prev.toAQL() + ' ' : '') +
    'SORT ' +
    args.join(', ')
  );
};

function LimitExpression(prev, offset, count) {
  if (count === undefined) {
    count = offset;
    offset = undefined;
  }
  this.prev = prev;
  this.offset = offset === undefined ? undefined : autoCastToken(offset);
  this.count = autoCastToken(count);
}
LimitExpression.prototype = new PartialStatement();
LimitExpression.prototype.constructor = LimitExpression;
LimitExpression.prototype.toAQL = function () {
  return (
    (this.prev ? this.prev.toAQL() + ' ' : '') +
    'LIMIT ' + (
      this.offset === undefined ?
      wrapAQL(this.count) :
      wrapAQL(this.offset) + ', ' + wrapAQL(this.count)
    )
  );
};

function Statement() {}
Statement.prototype = new Expression();
Statement.prototype.constructor = Statement;

function ReturnExpression(prev, value) {
  this.prev = prev;
  this.value = autoCastToken(value);
}
ReturnExpression.prototype = new Statement();
ReturnExpression.prototype.constructor = ReturnExpression;
ReturnExpression.prototype.toAQL = function () {
  return (
    (this.prev ? this.prev.toAQL() + ' ' : '') +
    'RETURN ' + wrapAQL(this.value)
  );
};

function RemoveExpression(prev, expr, collection) {
  this.prev = prev;
  this.expr = autoCastToken(expr);
  this.collection = new Identifier(collection);
}
RemoveExpression.prototype = new Statement();
RemoveExpression.prototype.constructor = RemoveExpression;
RemoveExpression.prototype.options = function (opts) {
  return new RemoveExpressionWithOptions(this.prev, this.expr, this.collection, opts);
};
RemoveExpression.prototype.toAQL = function () {
  return (
    (this.prev ? this.prev.toAQL() + ' ' : '') +
    'REMOVE ' + wrapAQL(this.expr) +
    ' IN ' + wrapAQL(this.collection)
  );
};

function RemoveExpressionWithOptions(prev, expr, collection, opts) {
  this.prev = prev;
  this.expr = autoCastToken(expr);
  this.collection = new Identifier(collection);
  this.opts = autoCastToken(opts);
}
RemoveExpressionWithOptions.prototype = new Statement();
RemoveExpressionWithOptions.prototype.constructor = RemoveExpressionWithOptions;
RemoveExpressionWithOptions.prototype.toAQL = function () {
  return (
    (this.prev ? this.prev.toAQL() + ' ' : '') +
    'REMOVE ' + wrapAQL(this.expr) +
    ' IN ' + wrapAQL(this.collection) +
    ' OPTIONS ' + wrapAQL(this.opts)
  );
};

function InsertExpression(prev, expr, collection) {
  this.prev = prev;
  this.expr = autoCastToken(expr);
  this.collection = new Identifier(collection);
}
InsertExpression.prototype = new Statement();
InsertExpression.prototype.constructor = InsertExpression;
InsertExpression.prototype.options = function (opts) {
  return new InsertExpressionWithOptions(this.prev, this.expr, this.collection, opts);
};
InsertExpression.prototype.toAQL = function () {
  return (
    (this.prev ? this.prev.toAQL() + ' ' : '') +
    'INSERT ' + wrapAQL(this.expr) +
    ' INTO ' + wrapAQL(this.collection)
  );
};

function InsertExpressionWithOptions(prev, expr, collection, opts) {
  this.prev = prev;
  this.expr = autoCastToken(expr);
  this.collection = new Identifier(collection);
  this.opts = autoCastToken(opts);
}
InsertExpressionWithOptions.prototype = new Statement();
InsertExpressionWithOptions.prototype.constructor = InsertExpressionWithOptions;
InsertExpressionWithOptions.prototype.toAQL = function () {
  return (
    (this.prev ? this.prev.toAQL() + ' ' : '') +
    'INSERT ' + wrapAQL(this.expr) +
    ' INTO ' + wrapAQL(this.collection) +
    ' OPTIONS ' + wrapAQL(this.opts)
  );
};

function UpdateExpression(prev, expr, withExpr, collection) {
  this.prev = prev;
  this.expr = autoCastToken(expr);
  this.withExpr = withExpr === undefined ? undefined : autoCastToken(withExpr);
  this.collection = new Identifier(collection);
}
UpdateExpression.prototype = new Statement();
UpdateExpression.prototype.constructor = UpdateExpression;
UpdateExpression.prototype.options = function (opts) {
  return new UpdateExpressionWithOptions(this.prev, this.expr, this.withExpr, this.collection, opts);
};
UpdateExpression.prototype.toAQL = function () {
  return (
    (this.prev ? this.prev.toAQL() + ' ' : '') +
    'UPDATE ' + wrapAQL(this.expr) +
    (this.withExpr ? ' WITH ' + wrapAQL(this.withExpr) : '') +
    ' IN ' + wrapAQL(this.collection)
  );
};

function UpdateExpressionWithOptions(prev, expr, withExpr, collection, opts) {
  this.prev = prev;
  this.expr = autoCastToken(expr);
  this.withExpr = withExpr === undefined ? undefined : autoCastToken(withExpr);
  this.collection = new Identifier(collection);
  this.opts = autoCastToken(opts);
}
UpdateExpressionWithOptions.prototype = new Statement();
UpdateExpressionWithOptions.prototype.constructor = UpdateExpressionWithOptions;
UpdateExpressionWithOptions.prototype.toAQL = function () {
  return (
    (this.prev ? this.prev.toAQL() + ' ' : '') +
    'UPDATE ' + wrapAQL(this.expr) +
    (this.withExpr ? ' WITH ' + wrapAQL(this.withExpr) : '') +
    ' IN ' + wrapAQL(this.collection) +
    ' OPTIONS ' + wrapAQL(this.opts)
  );
};

function ReplaceExpression(prev, expr, withExpr, collection) {
  this.prev = prev;
  this.expr = autoCastToken(expr);
  this.withExpr = withExpr === undefined ? undefined : autoCastToken(withExpr);
  this.collection = new Identifier(collection);
}
ReplaceExpression.prototype = new Statement();
ReplaceExpression.prototype.constructor = ReplaceExpression;
ReplaceExpression.prototype.options = function (opts) {
  return new ReplaceExpressionWithOptions(this.prev, this.expr, this.withExpr, this.collection, opts);
};
ReplaceExpression.prototype.toAQL = function () {
  return (
    (this.prev ? this.prev.toAQL() + ' ' : '') +
    'REPLACE ' + wrapAQL(this.expr) +
    (this.withExpr ? ' WITH ' + wrapAQL(this.withExpr) : '') +
    ' IN ' + wrapAQL(this.collection)
  );
};

function ReplaceExpressionWithOptions(prev, expr, withExpr, collection, opts) {
  this.prev = prev;
  this.expr = autoCastToken(expr);
  this.withExpr = withExpr === undefined ? undefined : autoCastToken(withExpr);
  this.collection = new Identifier(collection);
  this.opts = autoCastToken(opts);
}
ReplaceExpressionWithOptions.prototype = new Statement();
ReplaceExpressionWithOptions.prototype.constructor = ReplaceExpressionWithOptions;
ReplaceExpressionWithOptions.prototype.toAQL = function () {
  return (
    (this.prev ? this.prev.toAQL() + ' ' : '') +
    'REPLACE ' + wrapAQL(this.expr) +
    (this.withExpr ? ' WITH ' + wrapAQL(this.withExpr) : '') +
    ' IN ' + wrapAQL(this.collection) +
    ' OPTIONS ' + wrapAQL(this.opts)
  );
};

exports.autoCastToken = autoCastToken;
exports.RawExpression = RawExpression;
exports.NullLiteral = NullLiteral;
exports.BooleanLiteral = BooleanLiteral;
exports.NumberLiteral = NumberLiteral;
exports.IntegerLiteral = IntegerLiteral;
exports.StringLiteral = StringLiteral;
exports.ListLiteral = ListLiteral;
exports.ObjectLiteral = ObjectLiteral;
exports.RangeExpression = RangeExpression;
exports.PropertyAccess = PropertyAccess;
exports.Keyword = Keyword;
exports.Identifier = Identifier;
exports.SimpleReference = SimpleReference;
exports.UnaryOperation = UnaryOperation;
exports.BinaryOperation = BinaryOperation;
exports.TernaryOperation = TernaryOperation;
exports.NAryOperation = NAryOperation;
exports.FunctionCall = FunctionCall;
exports.ForExpression = ForExpression;
exports.FilterExpression = FilterExpression;
exports.LetExpression = LetExpression;
exports.CollectExpression = CollectExpression;
exports.SortExpression = SortExpression;
exports.LimitExpression = LimitExpression;
exports.ReturnExpression = ReturnExpression;
exports.RemoveExpression = RemoveExpression;
exports.InsertExpression = InsertExpression;
exports.UpdateExpression = UpdateExpression;
exports.ReplaceExpression = ReplaceExpression;

exports._Expression = Expression;
exports._Operation = Operation;
exports._Statement = Statement;
exports._PartialStatement = PartialStatement;
exports._Definitions = Definitions;
exports._CollectIntoExpression = CollectIntoExpression;
exports._RemoveExpressionWithOptions = RemoveExpressionWithOptions;
exports._InsertExpressionWithOptions = InsertExpressionWithOptions;
exports._UpdateExpressionWithOptions = UpdateExpressionWithOptions;
exports._ReplaceExpressionWithOptions = ReplaceExpressionWithOptions;
