/*jshint strict: false, maxlen: 500 */
/*global require, assertTrue, assertEqual, AQL_EXPLAIN */

////////////////////////////////////////////////////////////////////////////////
/// @brief tests for optimizer rules
///
/// @file
///
/// DISCLAIMER
///
/// Copyright 2010-2014 triagens GmbH, Cologne, Germany
///
/// Licensed under the Apache License, Version 2.0 (the "License");
/// you may not use this file except in compliance with the License.
/// You may obtain a copy of the License at
///
///     http://www.apache.org/licenses/LICENSE-2.0
///
/// Unless required by applicable law or agreed to in writing, software
/// distributed under the License is distributed on an "AS IS" BASIS,
/// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
/// See the License for the specific language governing permissions and
/// limitations under the License.
///
/// Copyright holder is triAGENS GmbH, Cologne, Germany
///
/// @author 
/// @author Copyright 2014, triAGENS GmbH, Cologne, Germany
////////////////////////////////////////////////////////////////////////////////

var db = require("org/arangodb").db;
var jsunity = require("jsunity");
var helper = require("org/arangodb/aql-helper");

////////////////////////////////////////////////////////////////////////////////
/// @brief test suite
////////////////////////////////////////////////////////////////////////////////

function optimizerRuleTestSuite () {
  var ruleName = "remove-unnecessary-remote-scatter";
  // various choices to control the optimizer: 
  var rulesNone        = { optimizer: { rules: [ "-all" ] } };
  var rulesAll         = { optimizer: { rules: [ "+all" ] } };
  var thisRuleEnabled  = { optimizer: { rules: [ "-all", "+" + ruleName ] } };
  var thisRuleDisabled = { optimizer: { rules: [ "+all", "-" + ruleName ] } };

  var cn1 = "UnitTestsAqlOptimizerRuleRemoveUnnecessaryRemoteScatter1";
  var cn2 = "UnitTestsAqlOptimizerRuleRemoveUnnecessaryRemoteScatter2";
  var c1, c2;
  
  var explain = function (result) {
    return helper.getCompactPlan(result).map(function(node) 
        { return node.type; });
  };

  return {

    ////////////////////////////////////////////////////////////////////////////////
    /// @brief set up
    ////////////////////////////////////////////////////////////////////////////////

    setUp : function () {
      var i;
      db._drop(cn1);
      db._drop(cn2);
      c1 = db._create(cn1, {numberOfShards:9});
      c2 = db._create(cn2);
      for (i = 0; i < 10; i++){ 
          c1.insert({Hallo1:i});
          c2.insert({Hallo2:i});
      }
    },

    ////////////////////////////////////////////////////////////////////////////////
    /// @brief tear down
    ////////////////////////////////////////////////////////////////////////////////

    tearDown : function () {
      db._drop(cn1);
      db._drop(cn2);
    },

    ////////////////////////////////////////////////////////////////////////////////
    /// @brief test that rule does not fire when all rules are disabled 
    ////////////////////////////////////////////////////////////////////////////////

    testRulesNone : function () {
      var queries = [ 
        "FOR d IN " + cn1 + " RETURN d",
        "LET A=1 LET B=2 FOR d IN " + cn1 + " RETURN d",
        "LET A=1 LET B=2 FOR d IN " + cn1 + " LIMIT 10 RETURN d",
        "FOR e in " + cn1 + " FOR d IN " + cn1 + " LIMIT 10 RETURN d"
      ];
      queries.forEach(function(query) {
        var result = AQL_EXPLAIN(query, { }, rulesNone);
        assertEqual([ "scatter-in-cluster" ], result.plan.rules, query);

      });
    },

    /* // this currently crashes the cluster...
     * testRulesNoneSerial : function () {
      var queries = [ 
        "FOR d IN " + cn1 + " RETURN d",
        "LET A=1 LET B=2 FOR d IN " + cn1 + " RETURN d",
        "LET A=1 LET B=2 FOR d IN " + cn1 + " LIMIT 10 RETURN d",
        "FOR e in " + cn1 + " FOR d IN " + cn1 + " LIMIT 10 RETURN d"
      ];

      var opts = _.clone(rulesNone);
      opts.allPlans = true;
      opts.verbosePlans = true;

      queries.forEach(function(query) {
        var plans = AQL_EXPLAIN(query, { }, opts).plans;
        plans.forEach(function(plan) {
          var result = AQL_EXECUTEJSON(plan, rulesNone).json;
          assertEqual(query[1], result, query[0]);
        });
      });
    },*/

    ////////////////////////////////////////////////////////////////////////////////
    /// @brief test that rule does not fire when it is disabled but no other rule is
    ////////////////////////////////////////////////////////////////////////////////
    
    testThisRuleDisabled : function () {
      var queries = [ 
        "FOR d IN " + cn1 + " RETURN d",
        "LET A=1 LET B=2 FOR d IN " + cn1 + " RETURN d",
        "LET A=1 LET B=2 FOR d IN " + cn1 + " LIMIT 10 RETURN d",
        "FOR e in " + cn1 + " FOR d IN " + cn1 + " LIMIT 10 RETURN d"
      ];

      queries.forEach(function(query) {
        var result = AQL_EXPLAIN(query, { }, thisRuleDisabled);
        assertTrue(result.plan.rules.indexOf(ruleName) === -1, query);
      });
    },

    testThisRuleEnabled : function () {
      var queries = [ 
        [ "FOR d IN " + cn1 + " RETURN d", 0],
        [ "LET A=1 LET B=2 FOR d IN " + cn1 + " RETURN d" , 1],
        [ "LET A=1 LET B=2 FOR d IN " + cn1 + " LIMIT 10 RETURN d", 2],
        [ "FOR e in " + cn1 + " FOR d IN " + cn1 + " LIMIT 10 RETURN d", 3]
      ];

      var expectedRules = [ "scatter-in-cluster", "remove-unnecessary-remote-scatter"];
      var expectedNodes = [ 
          ["SingletonNode", "EnumerateCollectionNode",
           "RemoteNode", "GatherNode", "ReturnNode"],
          ["SingletonNode", "CalculationNode", "CalculationNode",
           "EnumerateCollectionNode", "RemoteNode", "GatherNode", "ReturnNode"],
          ["SingletonNode", "CalculationNode", "CalculationNode",
           "EnumerateCollectionNode", "RemoteNode",
           "GatherNode", "LimitNode", "ReturnNode" ],
          ["SingletonNode", "EnumerateCollectionNode", "RemoteNode",
           "GatherNode", "ScatterNode", "RemoteNode", "EnumerateCollectionNode",
           "RemoteNode", "GatherNode", "LimitNode", "ReturnNode"] ];

      queries.forEach(function(query) {
        var result = AQL_EXPLAIN(query[0], { }, thisRuleEnabled);
        assertEqual(expectedRules, result.plan.rules, query);
        assertEqual(expectedNodes[query[1]], explain(result), query);
      });
    },

    ////////////////////////////////////////////////////////////////////////////////
    /// @brief test that rule has no effect
    ////////////////////////////////////////////////////////////////////////////////

    testRuleNoEffect : function () {
      var queries = [  "LIMIT 1 FOR d IN " + cn1 + " RETURN d",
                       "LIMIT 1 FOR d IN " + cn2 + " RETURN d" ];

      queries.forEach(function(query) {
        var result1 = AQL_EXPLAIN(query, { }, thisRuleEnabled);
        var result2 = AQL_EXPLAIN(query, { }, rulesAll);

        assertTrue(result1.plan.rules.indexOf(ruleName) === -1, query);
        assertTrue(result2.plan.rules.indexOf(ruleName) === -1, query);
      });
    },

    ////////////////////////////////////////////////////////////////////////////////
    /// @brief test that rule has an effect
    ////////////////////////////////////////////////////////////////////////////////

    testRuleHasEffect : function () {
      var queries = [ 
        "FOR d IN " + cn1 + " RETURN d",
        "FOR d IN " + cn2 + " RETURN d"
      ];

      queries.forEach(function(query) {
        var resultEnabled  = AQL_EXPLAIN(query, { }, thisRuleEnabled);
        var resultDisabled = AQL_EXPLAIN(query, { }, thisRuleDisabled);
        assertTrue(resultEnabled.plan.rules.indexOf(ruleName)  !== -1, query);
        assertTrue(resultDisabled.plan.rules.indexOf(ruleName) === -1, query);
        
        // the test doesn't run with the below . . .
        //resultDisabled = AQL_EXECUTE(query, { }, thisRuleDisabled).json;
        //resultEnabled  = AQL_EXECUTE(query, { }, thisRuleEnabled).json;
        //assertEqual(resultDisabled, resultEnabled, query);
      });
    },
  };
}

////////////////////////////////////////////////////////////////////////////////
/// @brief executes the test suite
////////////////////////////////////////////////////////////////////////////////

jsunity.run(optimizerRuleTestSuite);

return jsunity.done();

